/*
 * xfrm4_mode_beet.c - BEET mode encapsulation for IPv4.
 *
 * Copyright (c) 2006 Diego Beltrami <diego.beltrami@gmail.com>
 *                    Miika Komu     <miika@iki.fi>
 *                    Herbert Xu     <herbert@gondor.apana.org.au>
 *                    Abhinav Pathak <abhinav.pathak@hiit.fi>
 *                    Jeff Ahrenholz <ahrenholz@gmail.com>
 *                    Joakim Koskela <jookos@gmail.com>
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/stringify.h>
#include <net/dst.h>
#include <net/ip.h>
#include <net/xfrm.h>
#include <net/inet_ecn.h>

/* Add encapsulation header.
 *
 * The top IP header will be constructed per draft-nikander-esp-beet-mode-06.txt.
 * The following fields in it shall be filled in by x->type->output:
 *      tot_len
 *      check
 *
 * On exit, skb->transport_header will be set to the start of the
 * payload to be processed by x->type->output and skb->network_header
 * will be set to the top IP header.
 */
static int xfrm4_beet_output(struct xfrm_state *x, struct sk_buff *skb)
{
	struct dst_entry *dst = skb->dst;
	struct iphdr *iphv4, *top_iphv4;
	int hdrlen;

	if (ip_hdr(skb)->version == 4) {
		int optlen;

		/* 4-4 */
		iphv4 = ip_hdr(skb);
		skb->transport_header = skb->network_header;

		hdrlen = 0;
		optlen = iphv4->ihl * 4 - sizeof(*iphv4);
		if (unlikely(optlen))
			hdrlen += IPV4_BEET_PHMAXLEN - (optlen & 4);

		skb_push(skb, x->props.header_len - IPV4_BEET_PHMAXLEN + hdrlen);
		skb_reset_network_header(skb);
		top_iphv4 = ip_hdr(skb);
		skb->transport_header += sizeof(*iphv4) - hdrlen;

		memmove(top_iphv4, iphv4, sizeof(*iphv4));
		if (unlikely(optlen)) {
			struct ip_beet_phdr *ph;

			BUG_ON(optlen < 0);

			ph = (struct ip_beet_phdr *)skb_transport_header(skb);
			ph->padlen = 4 - (optlen & 4);
			ph->hdrlen = optlen / 8;
			ph->nexthdr = iphv4->protocol;
			if (ph->padlen)
				memset(ph + 1, IPOPT_NOP, ph->padlen);

			top_iphv4->protocol = IPPROTO_BEETPH;
			top_iphv4->ihl = 5;
		}

		top_iphv4->saddr = x->props.saddr.a4;
		top_iphv4->daddr = x->id.daddr.a4;
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
	} else if (ip_hdr(skb)->version == 6) {
		int delta = sizeof(struct ipv6hdr) - sizeof(struct iphdr);
		u8 protocol = ipv6_hdr(skb)->nexthdr;

		skb_set_transport_header(skb, sizeof(struct ipv6hdr));

		/* Inner = 6, Outer = 4 : changing the external IP hdr
		 * to the outer addresses
		 */
		hdrlen = x->props.header_len;

		skb_pull(skb, delta);
		skb_push(skb, hdrlen);

		skb_reset_network_header(skb);
		top_iphv4 = ip_hdr(skb);
		top_iphv4->ihl = 5;
		top_iphv4->version = 4;
		top_iphv4->id = 0;
		top_iphv4->frag_off = htons(IP_DF);
		top_iphv4->ttl = dst_metric(dst->child, RTAX_HOPLIMIT);
		top_iphv4->protocol = protocol;

		top_iphv4->saddr = x->props.saddr.a4;
		top_iphv4->daddr = x->id.daddr.a4;
		IPCB(skb)->flags = 0;
#endif
	}

	skb->protocol = htons(ETH_P_IP);
	memset(&(IPCB(skb)->opt), 0, sizeof(struct ip_options));
	return 0;
}

static int xfrm4_beet_input(struct xfrm_state *x, struct sk_buff *skb)
{
	struct iphdr *iph = ip_hdr(skb);
	int hops = iph->ttl;
	int phlen = 0;
	int optlen = 0;
	__u8 protocol = 0;
	int err = -EINVAL;

	protocol = iph->protocol;
	if (x->sel.family == AF_INET) {
		if (unlikely(protocol == IPPROTO_BEETPH)) {
			struct ip_beet_phdr *ph;

			if (!pskb_may_pull(skb, sizeof(*ph)))
				goto out;

			ph = (struct ip_beet_phdr *)(ipip_hdr(skb) + 1);

			phlen = sizeof(*ph) + ph->padlen;
			optlen = ph->hdrlen * 8 + (IPV4_BEET_PHMAXLEN - phlen);
			if (optlen < 0 || optlen & 3 || optlen > 250)
				goto out;

			if (!pskb_may_pull(skb, phlen + optlen))
				goto out;

			protocol = ph->nexthdr;
		}

		skb_push(skb, sizeof(*iph) - phlen);
		skb_reset_network_header(skb);
		memmove(skb_network_header(skb), iph, sizeof(*iph));

		iph = ip_hdr(skb);
		iph->ihl = (sizeof(*iph) + optlen) / 4;
		iph->tot_len = htons(skb->len);
		iph->daddr = x->sel.daddr.a4;
		iph->saddr = x->sel.saddr.a4;
		iph->protocol = protocol;

		ip_send_check(iph);
		if (ip_options_compile(NULL, skb))
			goto out;

		dst_release(skb->dst);
		skb->dst = NULL;

		pskb_pull(skb, sizeof(*iph) + optlen);
		skb_reset_transport_header(skb);
		err = 0;
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
	} else if (x->sel.family == AF_INET6) {
		int delta = sizeof(struct ipv6hdr) + skb->mac_len;
		struct ipv6hdr *ip6h;
		const unsigned char *old_mac;

		/* Here, the inner family is 6, therefore I have to
		 * substitute the IPhdr by enlarging it.
		 */
		if (skb_headroom(skb) < delta) {
			if (pskb_expand_head(skb, delta, 0, GFP_ATOMIC))
				goto out;
		}

		skb_push(skb, sizeof(*ip6h));

		/* mac might have references to ipv4. shouldn't matter */
		old_mac = skb_mac_header(skb);
		skb_set_mac_header(skb, -skb->mac_len);
		memmove(skb_mac_header(skb), old_mac, skb->mac_len);

		skb_reset_network_header(skb);
		skb_reset_transport_header(skb);

		ip6h = ipv6_hdr(skb);
		memset(ip6h->flow_lbl, 0, sizeof(ip6h->flow_lbl));
		ip6h->version = 6;
		ip6h->priority = 0;
		ip6h->nexthdr = protocol;
		ip6h->hop_limit = hops;
		ip6h->payload_len = htons(skb->len - sizeof(*ip6h));
		ipv6_addr_copy(&ip6h->daddr,
			       (struct in6_addr *)&x->sel.daddr.a6);
		ipv6_addr_copy(&ip6h->saddr,
			       (struct in6_addr *)&x->sel.saddr.a6);
		skb->protocol = htons(ETH_P_IPV6);
		err = 0;
#endif
	}
out:
	return err;
}

static struct xfrm_mode xfrm4_beet_mode = {
	.input = xfrm4_beet_input,
	.output = xfrm4_beet_output,
	.owner = THIS_MODULE,
	.encap = XFRM_MODE_BEET,
};

static int __init xfrm4_beet_init(void)
{
	return xfrm_register_mode(&xfrm4_beet_mode, AF_INET);
}

static void __exit xfrm4_beet_exit(void)
{
	int err;

	err = xfrm_unregister_mode(&xfrm4_beet_mode, AF_INET);
	BUG_ON(err);
}

module_init(xfrm4_beet_init);
module_exit(xfrm4_beet_exit);
MODULE_LICENSE("GPL");
MODULE_ALIAS_XFRM_MODE(AF_INET, XFRM_MODE_BEET);
