/*
 * xfrm6_mode_beet.c - BEET mode encapsulation for IPv6.
 *
 * Copyright (c) 2006 Diego Beltrami <diego.beltrami@gmail.com>
 *                    Miika Komu     <miika@iki.fi>
 *                    Herbert Xu     <herbert@gondor.apana.org.au>
 *                    Abhinav Pathak <abhinav.pathak@hiit.fi>
 *                    Jeff Ahrenholz <ahrenholz@gmail.com>
 *                    Joakim Koskela <jookos@gmail.com>
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/stringify.h>
#include <net/dsfield.h>
#include <net/dst.h>
#include <net/inet_ecn.h>
#include <net/ipv6.h>
#include <net/ip.h>
#include <net/xfrm.h>

/* Add encapsulation header.
 *
 * The top IP header will be constructed per draft-nikander-esp-beet-mode-06.txt.
 * The following fields in it shall be filled in by x->type->output:
 *	payload_len
 *
 * On exit, skb->h will be set to the start of the encapsulation header to be
 * filled in by x->type->output and skb->nh will be set to the nextheader field
 * of the extension header directly preceding the encapsulation header, or in
 * its absence, that of the top IP header.  The value of skb->data will always
 * point to the top IP header.
 */
static int xfrm6_beet_output(struct xfrm_state *x, struct sk_buff *skb)
{
	struct dst_entry *dst = skb->dst;
	struct iphdr *iphv4;
	struct ipv6hdr *iphv6, *top_iphv6;
	int hdr_len;

	if (ip_hdr(skb)->version == 6) {
		u8 *prevhdr;

		skb_push(skb, x->props.header_len);
		iphv6 = ipv6_hdr(skb);

		hdr_len = ip6_find_1stfragopt(skb, &prevhdr);
		skb_reset_network_header(skb);

		top_iphv6 = ipv6_hdr(skb);
		memmove(top_iphv6, iphv6, hdr_len);
		ipv6_addr_copy(&top_iphv6->saddr, (struct in6_addr *)&x->props.saddr);
		ipv6_addr_copy(&top_iphv6->daddr, (struct in6_addr *)&x->id.daddr);
		skb->transport_header = skb->network_header + sizeof(struct ipv6hdr);
		skb->network_header += offsetof(struct ipv6hdr, nexthdr);
	} else if (ip_hdr(skb)->version == 4) {
		int flags, optlen, dsfield;
		u8 protocol;

		iphv4 = ip_hdr(skb);
		skb->transport_header = skb->network_header;

		hdr_len = 0;
		optlen = iphv4->ihl * 4 - sizeof(*iphv4);
		if (unlikely(optlen))
			hdr_len += IPV4_BEET_PHMAXLEN - (optlen & 4);

		skb_push(skb, x->props.header_len - IPV4_BEET_PHMAXLEN + hdr_len);
		skb->transport_header += sizeof(*iphv4) - hdr_len;
		if (unlikely(optlen)) {
			struct ip_beet_phdr *ph;

			BUG_ON(optlen < 0);

			ph = (struct ip_beet_phdr *)skb_transport_header(skb);
			ph->padlen = 4 - (optlen & 4);
			ph->hdrlen = optlen / 8;
			ph->nexthdr = iphv4->protocol;
			if (ph->padlen)
				memset(ph + 1, IPOPT_NOP, ph->padlen);

			protocol = IPPROTO_BEETPH;
		} else
			protocol = iphv4->protocol;

		skb_reset_network_header(skb);
		top_iphv6 = ipv6_hdr(skb);

		/* DS disclosed */
		top_iphv6->version = 6;
		top_iphv6->priority = 0;
		top_iphv6->flow_lbl[0] = 0;
		top_iphv6->flow_lbl[1] = 0;
		top_iphv6->flow_lbl[2] = 0;
		dsfield = ipv6_get_dsfield(top_iphv6);
		dsfield = INET_ECN_encapsulate(dsfield, dsfield);
		flags = x->props.flags;
		if (flags & XFRM_STATE_NOECN)
			dsfield &= ~INET_ECN_MASK;
		ipv6_change_dsfield(top_iphv6, 0, dsfield);

		top_iphv6->nexthdr = protocol;
		top_iphv6->hop_limit = dst_metric(dst->child, RTAX_HOPLIMIT);
		top_iphv6->payload_len = htons(skb->len - sizeof(struct ipv6hdr));
		ipv6_addr_copy(&top_iphv6->saddr, (struct in6_addr *)&x->props.saddr);
		ipv6_addr_copy(&top_iphv6->daddr, (struct in6_addr *)&x->id.daddr);
		skb->transport_header = skb->network_header + sizeof(struct ipv6hdr);
		skb->network_header += offsetof(struct ipv6hdr, nexthdr);
	}

	skb->protocol = htons(ETH_P_IPV6);
	return 0;
}


static int xfrm6_beet_input(struct xfrm_state *x, struct sk_buff *skb)
{
	struct ipv6hdr *ip6h;
	const unsigned char *old_mac;
	int size = ((x->sel.family == AF_INET) ?
		    sizeof(struct iphdr) :
		    sizeof(struct ipv6hdr));
	int err = -EINVAL;

	if (x->sel.family == AF_INET6) {
		if (skb_cloned(skb) &&
		    pskb_expand_head(skb, 0, 0, GFP_ATOMIC))
			goto out;

		ip6h = ipv6_hdr(skb);
		skb_set_network_header(skb, -size);
		memmove(skb_network_header(skb), ip6h, size);

		ip6h = ipv6_hdr(skb);
		ipv6_addr_copy(&ip6h->daddr, (struct in6_addr *) &x->sel.daddr.a6);
		ipv6_addr_copy(&ip6h->saddr, (struct in6_addr *) &x->sel.saddr.a6);

		ip6h->payload_len = htons(skb->len);
		skb_reset_transport_header(skb);
		skb->protocol = htons(ETH_P_IPV6);
		err = 0;
	} else {
		__u8 proto = ipv6_hdr(skb)->nexthdr;
		__u8 hops = ipv6_hdr(skb)->hop_limit;
		int phlen = 0;
		int optlen = 0;
		struct iphdr *iph;

		/* Inner = IPv4, therefore the IPhdr must be shrunk */
		/* Inner = 4, Outer = 6 */
		if (unlikely(proto == IPPROTO_BEETPH)) {
			struct ip_beet_phdr *ph = (struct ip_beet_phdr *)
				(struct ip_beet_phdr *)skb->data;

			if (!pskb_may_pull(skb, sizeof(*ph)))
				goto out;

			phlen = sizeof(*ph) + ph->padlen;
			optlen = ph->hdrlen * 8 + (IPV4_BEET_PHMAXLEN - phlen);
			if (optlen < 0 || optlen & 3 || optlen > 250)
				goto out;

			if (!pskb_may_pull(skb, phlen + optlen))
				goto out;
			proto = ph->nexthdr;
		}

		if (skb_cloned(skb) &&
		    pskb_expand_head(skb, 0, 0, GFP_ATOMIC))
			goto out;

		skb_push(skb, size - phlen);
		skb_reset_network_header(skb);
		skb_reset_transport_header(skb);

		/* mac might have references to ipv6. shouldn't matter */
		old_mac = skb_mac_header(skb);
		skb_set_mac_header(skb, -skb->mac_len);
		memmove(skb_mac_header(skb), old_mac, skb->mac_len);

		iph = ip_hdr(skb);
		iph->ihl = (sizeof(*iph) + optlen) / 4;
		iph->version = 4;
		iph->tos = 0;
		iph->id = 0;
		iph->frag_off = 0;
		iph->ttl = hops;
		iph->protocol = proto;
		iph->daddr = x->sel.daddr.a4;
		iph->saddr = x->sel.saddr.a4;
		iph->tot_len = htons(skb->len);
		ip_send_check(iph);
		skb->protocol = htons(ETH_P_IP);
		err = 0;
	}
 out:
	return err;
}

static struct xfrm_mode xfrm6_beet_mode = {
	.input = xfrm6_beet_input,
	.output = xfrm6_beet_output,
	.owner = THIS_MODULE,
	.encap = XFRM_MODE_BEET,
};

static int __init xfrm6_beet_init(void)
{
	return xfrm_register_mode(&xfrm6_beet_mode, AF_INET6);
}

static void __exit xfrm6_beet_exit(void)
{
	int err;

	err = xfrm_unregister_mode(&xfrm6_beet_mode, AF_INET6);
	BUG_ON(err);
}

module_init(xfrm6_beet_init);
module_exit(xfrm6_beet_exit);
MODULE_LICENSE("GPL");
MODULE_ALIAS_XFRM_MODE(AF_INET6, XFRM_MODE_BEET);
